import socket
import time

import network
import uasyncio

from .core import eISCPPacket, parse_info
from .eiscp import eISCP


def get_current_ip_address() -> str:
    wlan_network = network.WLAN(network.STA_IF)
    if not wlan_network.isconnected():
        raise OSError("Must connect to network first")
    connection_details = wlan_network.ifconfig()
    return connection_details[0]


def get_current_broadcast_address() -> str:
    wlan_network = network.WLAN(network.STA_IF)
    if not wlan_network.isconnected():
        raise OSError("Must connect to network first")
    connection_details = wlan_network.ifconfig()
    ip_address = connection_details[0]
    netmask = connection_details[1]
    ip_address_segments = (int(item) for item in ip_address.split("."))
    netmask_segments = (int(item) for item in netmask.split("."))
    broadcast_segments = []
    for (ip_address_segment, netmask_segment) in zip(ip_address_segments, netmask_segments):
        if int(netmask_segment) == 255:
            broadcast_segments.append(ip_address_segment)
        else:
            network_segment = ip_address_segment & netmask_segment
            broadcast_segments.append(network_segment + (255 - netmask_segment))
    return ".".join(str(item) for item in broadcast_segments)


discovery_lock = uasyncio.Lock()


async def discover(timeout: int = 5, clazz=None):
    onkyo_magic = eISCPPacket("!xECNQSTN").get_raw()
    pioneer_magic = eISCPPacket("!pECNQSTN").get_raw()
    found_receivers = {}
    broadcast_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
    broadcast_socket.setblocking(0)
    broadcast_address = get_current_broadcast_address()
    own_address = get_current_ip_address()

    async with discovery_lock:
        try:
            broadcast_socket.bind(
                (own_address, eISCP.ONKYO_PORT)
            )  # The port doesn't matter. It is "0" in the original implementation. MicroPython doesn't support this.
            broadcast_socket.sendto(onkyo_magic, (broadcast_address, eISCP.ONKYO_PORT))
            broadcast_socket.sendto(pioneer_magic, (broadcast_address, eISCP.ONKYO_PORT))
            start = time.ticks_ms()
            while True:
                ready = uasyncio.select.select([broadcast_socket], [], [], 0.01)
                if not ready[0]:
                    await uasyncio.sleep_ms(100)
                else:
                    data, addr = broadcast_socket.recvfrom(1024)
                    info = parse_info(data)
                    receiver = (clazz or eISCP)(addr[0], int(info["iscp_port"]))
                    receiver.info = info
                    found_receivers[info["identifier"]] = receiver

                if start + timeout * 1000 < time.ticks_ms():
                    break

        finally:
            broadcast_socket.close()
    return list(found_receivers.values())



import re
import struct
import time
from collections import namedtuple


class ISCPMessage(object):
    """Deals with formatting and parsing data wrapped in an ISCP
    containers. The docs say:
        ISCP (Integra Serial Control Protocol) consists of three
        command characters and parameter character(s) of variable
        length.
    It seems this was the original protocol used for communicating
    via a serial cable.
    """

    def __init__(self, data: str):
        self.data = data

    def __str__(self) -> str:
        # ! = start character
        # 1 = destination unit type, 1 means receiver
        # End character may be CR, LF or CR+LF, according to doc
        return "!1{}\r".format(self.data)

    @classmethod
    def parse(self, data: bytes) -> bytes:
        EOF = "\x1a"
        TERMINATORS = ["\n", "\r"]
        assert data[:2] == "!1"
        eof_offset = -1
        # EOF can be followed by CR/LF/CR+LF
        if data[eof_offset] in TERMINATORS:
            eof_offset -= 1
            if data[eof_offset] in TERMINATORS:
                eof_offset -= 1
        assert data[eof_offset] == EOF
        return data[2:eof_offset]


class eISCPPacket(object):
    """For communicating over Ethernet, traditional ISCP messages are
    wrapped inside an eISCP package.
    """

    header = namedtuple("header", ("magic", "header_size", "data_size", "version", "reserved"))

    def __init__(self, iscp_message: "Union[bytes, str]"):
        iscp_message = str(iscp_message)
        # We attach data separately, because Python's struct module does
        # not support variable length strings,
        header = struct.pack(
            "!4sIIb3s",
            b"ISCP",  # magic
            16,  # header size (16 bytes)
            len(iscp_message),  # data size
            0x01,  # version
            b"\x00\x00\x00",  # reserved
        )

        if isinstance(iscp_message, str):
            iscp_message = iscp_message.encode("utf-8")

        self._bytes = header + iscp_message
        # __new__, string subclass?

    def __str__(self):
        return self._bytes.decode("utf-8")

    def get_raw(self):
        return self._bytes

    @classmethod
    def parse(cls, bytes):
        """Parse the eISCP package given by ``bytes``."""
        h = cls.parse_header(bytes[:16])
        start_data = h.header_size
        end_data = h.header_size + h.data_size
        data = bytes[start_data:end_data].decode()
        assert len(data) == h.data_size
        return data

    @classmethod
    def parse_header(self, bytes):
        """Parse the header of an eISCP package.
        This is useful when reading data in a streaming fashion,
        because you can subsequently know the number of bytes to
        expect in the packet.
        """
        # A header is always 16 bytes in length
        assert len(bytes) == 16

        # Parse the header
        magic, header_size, data_size, version, reserved = struct.unpack("!4sIIb3s", bytes)

        magic = magic.decode()
        reserved = reserved.decode()

        # Strangly, the header contains a header_size field.
        assert magic == "ISCP"
        assert header_size == 16

        return eISCPPacket.header(magic, header_size, data_size, version, reserved)


def parse_info(data: bytes) -> "Dict[str, str]":
    response = eISCPPacket.parse(data)
    # Return string looks something like this:
    # !1ECNTX-NR609/60128/DX
    matched = re.match(r"""!(\d)ECN([^/]*)/(\d\d\d\d\d)/(\w\w)/(.*)""", response.strip())
    if matched is None:
        return None

    identifier = matched.group(5)
    if len(identifier) > 12:
        identifier = identifier[0:12]

    info = {
        "device_category": matched.group(1),
        "model_name": matched.group(2),
        "iscp_port": matched.group(3),
        "area_code": matched.group(4),
        "identifier": identifier,
    }
    return info


def command_to_packet(command: str) -> bytes:
    """
    Convert an ascii command like (PVR00) to the binary data we
    need to send to the receiver.
    """
    return eISCPPacket(ISCPMessage(command)).get_raw()


async def filter_for_message(getter_func: "Callable[[], Optional[str]]", msg: str) -> str:
    """Helper that calls ``getter_func`` until a matching message
    is found, or the timeout occurs. Matching means the same commands
    group, i.e. for sent message MVLUP we would accept MVL13
    in response."""
    start = time.time()
    while True:
        candidate = await getter_func(0.05)
        # It seems ISCP commands are always three characters.
        if candidate and candidate[:3] == msg[:3]:
            return candidate

        # exception for HDMI-CEC commands (CTV) since they don't provide any response/confirmation
        if "CTV" in msg[:3]:
            return msg

        # The protocol docs claim that a response  should arrive
        # within *50ms or the communication has failed*. In my tests,
        # however, the interval needed to be at least 200ms before
        # I managed to see any response, and only after 300ms
        # reproducably, so use a generous timeout.
        if time.time() - start > 1.5:
            raise ValueError("Timeout waiting for response.")


def command_to_iscp(command: str, argument: str) -> str:
    return "{}{}".format(command, argument)


def iscp_to_command(iscp_message: str) -> "Tuple[str, str]":
    command, args = iscp_message[:3], iscp_message[3:]
    return command, args



import select
import socket
import time

import uasyncio

from .core import (
    ISCPMessage,
    command_to_iscp,
    command_to_packet,
    eISCPPacket,
    filter_for_message,
    iscp_to_command,
    parse_info,
)


class eISCP:
    """Implements the eISCP interface to Onkyo receivers.
    This uses a blocking interface. The remote end will regularily
    send unsolicited status updates. You need to manually call
    ``get_message`` to query those.
    You may want to look at the :meth:`Receiver` class instead, which
    uses a background thread.
    """

    ONKYO_PORT = 60128
    CONNECT_TIMEOUT = 5

    def __init__(self, host: str, port: int = 60128, debug: bool = False):
        self.host = host
        self.port = port
        self._info = None
        self.debug = debug

        self.async_command_socket = None
        self.command_socket = None

    @property
    def model_name(self) -> str:
        if self.info and self.info.get("model_name"):
            return self.info["model_name"]
        else:
            return "unknown-model"

    @property
    def identifier(self) -> str:
        if self.info and self.info.get("identifier"):
            return self.info["identifier"]
        else:
            return "no-id"

    def __repr__(self) -> str:
        if self.info and self.info.get("model_name"):
            model = self.info["model_name"]
        else:
            model = "unknown"
        string = "<{}({}) {}:{}>".format(self.__class__.__name__, model, self.host, self.port)
        return string

    @property
    def info(self) -> "Dict[str, str]":
        if not self._info:
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
            sock.setblocking(0)
            sock.bind(("0.0.0.0", self.ONKYO_PORT))
            sock.sendto(eISCPPacket("!xECNQSTN").get_raw(), (self.host, self.port))

            ready = select.select([sock], [], [], 0.1)
            if ready[0]:
                data = sock.recv(1024)
                self._info = parse_info(data)
            sock.close()
        return self._info

    @info.setter
    def info(self, value: "Dict[str, str]") -> None:
        self._info = value

    def _ensure_socket_connected(self) -> None:
        if self.command_socket is None:
            command_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            command_socket.settimeout(self.CONNECT_TIMEOUT)
            command_socket.connect((self.host, self.port))
            command_socket.setblocking(0)
            self.command_socket = command_socket
            self.async_command_socket = uasyncio.StreamWriter(command_socket)

    def disconnect(self) -> None:
        try:
            self.command_socket.close()
        except Exception:
            pass
        self.command_socket = None

    def __enter__(self):
        self._ensure_socket_connected()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.disconnect()

    async def send(self, iscp_message: str) -> None:
        """Send a low-level ISCP message, like ``MVL50``.
        This does not return anything, nor does it wait for a response
        from the receiver. You can query responses via :meth:`get`,
        or use :meth:`raw` to send a message and waiting for one.
        """
        self._ensure_socket_connected()
        self.async_command_socket.write(command_to_packet(iscp_message))
        await self.async_command_socket.drain()

    async def get(self, timeout: int = 0.2) -> bytes:
        """Return the next message sent by the receiver, or, after
        ``timeout`` has passed, return ``None``.
        """
        self._ensure_socket_connected()

        start = time.ticks_ms()
        header_bytes = b""
        while start + int(timeout * 1000) > time.ticks_ms() and len(header_bytes) < 16:
            self.dprint("Reading from socket for header")
            result_payload = self.command_socket.read(16 - len(header_bytes))
            if result_payload:
                header_bytes += result_payload
                self.dprint("Got {} header bytes".format(len(header_bytes)))

            if len(header_bytes) < 16:
                self.dprint("Complete header not retrieved, waiting for buffer to keep up")
                await uasyncio.sleep_ms(1)

        if len(header_bytes) < 16:
            return None

        header = eISCPPacket.parse_header(header_bytes)
        self.dprint("Found ISCP header {}".format(header))
        body = b""
        start = time.ticks_ms()
        while len(body) < header.data_size:
            result = self.command_socket.read(header.data_size - len(body))
            if result:
                body += result
            if start + timeout * 1000 < time.ticks_ms():
                return None
            elif len(body) < header.data_size:
                await uasyncio.sleep_ms(1)
        message = ISCPMessage.parse(body.decode())
        self.dprint("Identified ISCP response: {}".format(message))
        return message

    async def raw(self, iscp_message, expect_response: bool = True):
        """Send a low-level ISCP message, like ``MVL50``, and wait
        for a response.
        While the protocol is designed to acknowledge each message with
        a response, there is no fool-proof way to differentiate those
        from unsolicited status updates, though we'll do our best to
        try. Generally, this won't be an issue, though in theory the
        response this function returns to you sending ``SLI05`` may be
        an ``SLI06`` update from another controller.
        It'd be preferable to design your app in a way where you are
        processing all incoming messages the same way, regardless of
        their origin.
        """
        while await self.get(False):
            # Clear all incoming messages. If not yet queried,
            # they are lost. This is so that we can find the real
            # response to our sent command later.
            pass
        await self.send(iscp_message)
        if expect_response:
            return await filter_for_message(self.get, iscp_message)
        else:
            while await self.get(False):
                # Drain anything in the buffer
                pass
            return None

    async def command(self, command: str, argument: str, expect_response: bool = True) -> "Optional[Tuple[str, str]]":
        """Send a high-level command to the receiver, return the
        receiver's response formatted has a command.
        This is basically a helper that combines :meth:`raw`,
        :func:`command_to_iscp` and :func:`iscp_to_command`.
        """
        iscp_message = command_to_iscp(command, argument)
        response = await self.raw(iscp_message, expect_response)
        if response:
            return iscp_to_command(response)

        return None

    async def power_on(self) -> "Optional[Tuple[str, str]]":
        """Turn the receiver power on."""
        return await self.command("PWR", "01")

    async def power_off(self) -> "Optional[Tuple[str, str]]":
        """Turn the receiver power off."""
        return await self.command("PWR", "00")

    def dprint(self, item: str) -> None:
        if self.debug:
            print(item)
